from .base_translator import BaseTranslator
from .translated_object import TranslatedObject

DEFAULT_TRANSLATION_ENDPOINT: str = "https://translate.google.com/translate_a/single"
DEFAULT_TTS_ENDPOINT: str = "https://translate.google.com/translate_tts"


class Device:
    DEVICES: tuple[str, ...] = (
        "Linux; U; Android 10; Pixel 4",
        "Linux; U; Android 10; Pixel 4 XL",
        "Linux; U; Android 10; Pixel 4a",
        "Linux; U; Android 10; Pixel 4a XL",
        "Linux; U; Android 11; Pixel 4",
        "Linux; U; Android 11; Pixel 4 XL",
        "Linux; U; Android 11; Pixel 4a",
        "Linux; U; Android 11; Pixel 4a XL",
        "Linux; U; Android 11; Pixel 5",
        "Linux; U; Android 11; Pixel 5a",
        "Linux; U; Android 12; Pixel 4",
        "Linux; U; Android 12; Pixel 4 XL",
        "Linux; U; Android 12; Pixel 4a",
        "Linux; U; Android 12; Pixel 4a XL",
        "Linux; U; Android 12; Pixel 5",
        "Linux; U; Android 12; Pixel 5a",
        "Linux; U; Android 12; Pixel 6",
        "Linux; U; Android 12; Pixel 6 Pro",
    )

    __i = 0

    @classmethod
    def shift(cls) -> str:
        """Get next device string in rotation.

        Returns:
            str: Next device user agent string
        """
        cls.__i += 1
        cls.__i %= len(cls.DEVICES)
        return cls.DEVICES[cls.__i]


def get_base_headers() -> dict[str, str]:
    return {"User-Agent": f"GoogleTranslate/6.28.0.05.421483610 ({Device.shift()})"}


__all__ = [
    "BaseTranslator",
    "TranslatedObject",
    "DEFAULT_TRANSLATION_ENDPOINT",
    "DEFAULT_TTS_ENDPOINT",
    "Device",
    "get_base_headers",
]



import io
from collections.abc import Mapping
from typing import Any, Callable, Dict, List, Optional, Protocol, TypeVar, Union, overload

import httpx

from .exceptions import TranslationError
from .types import (
    DEFAULT_TRANSLATION_ENDPOINT,
    DEFAULT_TTS_ENDPOINT,
    BaseTranslator,
    TranslatedObject,
    get_base_headers,
)

T = TypeVar("T", str, List[str], Dict[Any, str], Mapping[Any, str])
K = TypeVar("K")


class AsyncBufferedIOBase(Protocol):
    async def write(self, data: bytes) -> int: ...
    async def close(self) -> None: ...


class Translator(BaseTranslator):
    def __init__(
        self,
        proxies: Optional[Dict[str, str]] = None,
        url: str = DEFAULT_TRANSLATION_ENDPOINT,
        tts_url: str = DEFAULT_TTS_ENDPOINT,
        headers: Optional[Union[Dict[str, str], Callable[[], Dict[str, str]]]] = None,
        **options: Any,
    ) -> None:
        """Initialize the translator.

        Args:
            proxies: Optional proxy configuration dictionary
            url: Translation API endpoint URL
            tts_url: Text-to-speech API endpoint URL
            headers: Custom headers or header generator function
            **options: Additional options passed to httpx.AsyncClient
        """
        self.url = url
        self.tts_url = tts_url
        self.proxies = proxies
        self.options = options
        self.headers = get_base_headers if headers is None else headers

    @overload
    async def translate(
        self,
        text: str,
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> TranslatedObject: ...

    @overload
    async def translate(
        self,
        text: List[str],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> List[TranslatedObject]: ...

    @overload
    async def translate(
        self,
        text: Dict[Any, str],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> Dict[str, TranslatedObject]: ...

    async def translate(
        self,
        text: Union[str, List[str], Dict[Any, str], Mapping[str, str]],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        # Literal["t", "at", "rm", "bd", "md", "ss", "ex", "rw", "dj"] = "t",
        # broken compatibility with python <3.8 :(
        dj: int = 1,
        **extra: Any,
    ) -> Union[TranslatedObject, List[TranslatedObject], Dict[str, TranslatedObject]]:
        """
        A function that translates text.

        Parameters:
            text: str = Text to translate.


            sourcelang: str = "text" original language, pass "auto" for auto-detection (default: "auto").


            targetlang: str = target language for translating "text" (default: "en").


            client: str = Google Translate client platform I guess, it's not well documented. (default: "gtx")
                (Will return the raw API json if "client" is not "gtx")


            dt: str = Specifies what to return to the API reply (default: "t").
                (Will return the raw API json if "dt" is not "t".
                t - Translation of source text;
                at - Alternate translations;
                rm - Transcription / transliteration of source and translated texts;
                bd - Dictionary, in case source text is one word
                     (you get translations with articles, reverse translations, etc.);
                md - Definitions of source text, if it's one word;
                ss - Synonyms of source text, if it's one word;
                ex - Examples;
                rw - See also list;
                dj - Json response with names. (dj=1).)

            dj: int = Probably specifies dictionary output if true, else list (default: 1).

            extra: dict = Extra params to be sent to api endpoint.

        Returns:
            TranslatedObject or JSON serialized object.
        """

        params = {
            k: v
            for k, v in {
                "client": client,
                "sl": sourcelang,
                "tl": targetlang,
                "dt": dt,
                "ie": "utf-8",
                "oe": "utf-8",
                "dj": dj,
                **extra,
            }.items()
            if v is not None
        }
        try:
            proxies: Dict[str, httpx.AsyncHTTPTransport] = {}
            if self.proxies:
                if proxies.get("https"):
                    proxies["https"] = httpx.AsyncHTTPTransport(proxy=self.proxies["https"])
                if proxies.get("http"):
                    proxies["http"] = httpx.AsyncHTTPTransport(proxy=self.proxies["http"])
                if proxies.get("socks5"):
                    proxies["socks5"] = httpx.AsyncHTTPTransport(proxy=self.proxies["socks5"])
                if proxies.get("socks5h"):
                    proxies["socks5h"] = httpx.AsyncHTTPTransport(proxy=self.proxies["socks5h"])

            async with httpx.AsyncClient(mounts=proxies, **self.options) as http_client:
                raw: Union[Mapping[str, Any], List[Any]] = (
                    (
                        await http_client.post(
                            self.url,
                            params={**params, "q": text},
                            headers=self.get_headers(),
                        )
                    ).json()
                    if isinstance(text, str)
                    else (
                        {
                            k: (
                                await http_client.post(
                                    self.url,
                                    params={**params, "q": v},
                                    headers=self.get_headers(),
                                )
                            ).json()
                            for k, v in text.items()
                        }
                        if isinstance(text, Mapping)
                        else [
                            (
                                await http_client.post(
                                    self.url,
                                    params={**params, "q": elem},
                                    headers=self.get_headers(),
                                )
                            ).json()
                            for elem in text
                        ]
                    )
                )
                await http_client.aclose()

            return self.check(raw=raw, client=client, dt=dt, text=text)
        except Exception as e:
            raise TranslationError(e) from None

    async def detect(self, text: Union[str, List[Any], Dict[Any, Any]]) -> Union[str, List[str], Dict[Any, str]]:
        if isinstance(text, str):
            return (await self(text)).lang
        elif isinstance(text, list):
            return [(await self(elem)).lang for elem in text]
        elif isinstance(text, dict):
            return {k: (await self(v)).lang for k, v in text.items()}
        else:
            raise ValueError("Language detection works only with str, list and dict")

    async def tts(
        self,
        text: Union[str, List[str], Dict[Any, str], Mapping[K, str]],
        file: Union[AsyncBufferedIOBase, io.BytesIO],
        targetlang: str = "en",
        client: str = "at",
        idx: int = 0,
        prev: str = "input",
        chunk_size: int = 1024,
        textlen: Optional[int] = None,
        **extra: Any,
    ) -> Union[AsyncBufferedIOBase, io.BytesIO]:
        """Generate text-to-speech audio.

        Args:
            text: Text to convert to speech
            file: Output file or buffer
            targetlang: Target language code
            client: API client identifier
            idx: TTS segment index
            prev: Previous segment identifier
            chunk_size: Download chunk size
            textlen: Override text length
            **extra: Additional TTS parameters

        Returns:
            The output file/buffer with audio data

        Raises:
            TranslationError: If TTS generation fails
            ValueError: If targetlang is invalid
        """
        if not isinstance(targetlang, str) or len(targetlang) != 2:
            raise ValueError("targetlang must be a 2-letter language code")
        params = self.parse_tts(
            client=client,
            targetlang=targetlang,
            idx=idx,
            prev=prev,
            text=text,
            textlen=textlen,
            extra=extra,
        )
        try:
            proxies: Dict[str, httpx.AsyncHTTPTransport] = {}
            if self.proxies:
                if proxies.get("https"):
                    proxies["https"] = httpx.AsyncHTTPTransport(proxy=self.proxies["https"])
                if proxies.get("http"):
                    proxies["http"] = httpx.AsyncHTTPTransport(proxy=self.proxies["http"])
                if proxies.get("socks5"):
                    proxies["socks5"] = httpx.AsyncHTTPTransport(proxy=self.proxies["socks5"])
                if proxies.get("socks5h"):
                    proxies["socks5h"] = httpx.AsyncHTTPTransport(proxy=self.proxies["socks5h"])

            async with httpx.AsyncClient(mounts=proxies, **self.options) as c:
                async with c.stream(
                    "GET",
                    url=self.tts_url,
                    params=params,
                    headers=self.get_headers(),
                ) as response:
                    if isinstance(file, io.BytesIO):
                        async for chunk in response.aiter_bytes(chunk_size=chunk_size):
                            file.write(chunk)
                    else:
                        async for chunk in response.aiter_bytes(chunk_size=chunk_size):
                            await file.write(chunk)
                await c.aclose()

            return file
        except Exception as e:
            raise TranslationError(e) from None

    __call__ = translate  # Backwards compatibility



from collections.abc import Mapping
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union

from .translated_object import TranslatedObject

K = TypeVar("K")
V = TypeVar("V")


class BaseTranslator:
    headers: Union[Dict[str, str], Callable[[], Dict[str, str]]]

    @staticmethod
    def parse(raw: Dict[str, Any], translated: bool = True) -> Union[TranslatedObject, Dict[str, Any]]:
        """Parse raw API response into TranslatedObject.

        Args:
            raw: Raw response from translation API
            translated: Whether to return TranslatedObject or dict

        Returns:
            Either TranslatedObject or raw dict based on translated parameter
        """
        if translated:
            return TranslatedObject.from_raw_response(raw)
        return raw

    def check(
        self,
        text: Union[str, Mapping[K, str], List[str]],
        raw: Union[Mapping[str, Any], List[Any]],
        client: str,
        dt: str,
    ) -> Union[TranslatedObject, Dict[K, TranslatedObject], List[TranslatedObject]]:
        """Check and validate translation API response.

        Args:
            text: Original input text
            raw: Raw API response data
            client: API client identifier
            dt: Response data type

        Returns:
            Parsed translation result(s)

        Raises:
            TranslationError: If response validation fails
        """
        """Check and parse API response based on input type.

        Args:
            text: Original input text
            raw: Raw API response
            client: API client type
            dt: Data type parameter

        Returns:
            Parsed translation result(s)
        """
        if client != "gtx" or dt != "t":
            return raw  # type: ignore

        if isinstance(text, str):
            return self.parse(raw)  # type: ignore
        elif isinstance(text, Mapping):
            return {k: self.parse(v) for k, v in raw.items()}  # type: ignore
        else:
            return [self.parse(elem) for elem in raw]  # type: ignore

    def get_headers(self) -> Dict[str, str]:
        return self.headers() if callable(self.headers) else self.headers

    @staticmethod
    def parse_tts(
        client: str,
        targetlang: str,
        idx: int,
        prev: str,
        text: Union[str, List[str], Dict[Any, str], Mapping[K, str]],
        textlen: Optional[int],
        extra: Dict[str, Any],
    ) -> Dict[str, Union[str, int]]:
        return {
            k: v
            for k, v in {
                "client": client,
                "tl": targetlang,
                "ie": "utf-8",
                "oe": "utf-8",
                "idx": idx,
                "prev": prev,
                "textlen": len(text) if textlen is None else textlen,
                "q": text,
                **extra,
            }.items()
            if v is not None
        }



"""Translation result object implementation."""

from dataclasses import dataclass
from typing import Any, Dict, List


@dataclass
class TranslatedObject:
    """A dataclass that holds translation results."""

    raw: Dict[str, Any]
    orig: str
    text: str
    orig_raw: List[str]
    text_raw: List[str]
    lang: str

    def __str__(self) -> str:
        """Get string representation with the translated text.

        Returns:
            str: The translated text
        """
        return self.text

    @classmethod
    def from_raw_response(cls, raw: Dict[str, Any]) -> "TranslatedObject":
        """Create TranslatedObject from raw API response.

        Args:
            raw: Raw response dictionary from the translation API

        Returns:
            TranslatedObject: Parsed translation result
        """
        return cls(
            raw=raw,
            orig=" ".join(s["orig"] for s in raw["sentences"] if "orig" in s),
            text=" ".join(s["trans"] for s in raw["sentences"] if "trans" in s),
            orig_raw=[s["orig"] for s in raw["sentences"] if "orig" in s],
            text_raw=[s["trans"] for s in raw["sentences"] if "trans" in s],
            lang=raw["src"],
        )



from collections.abc import Mapping
from typing import Any, BinaryIO, Callable, Dict, List, Optional, TypeVar, Union, overload

import httpx

from ..exceptions import TranslationError
from ..types import (
    DEFAULT_TRANSLATION_ENDPOINT,
    DEFAULT_TTS_ENDPOINT,
    BaseTranslator,
    TranslatedObject,
    get_base_headers,
)

T = TypeVar("T", str, List[str], Dict[Any, str], Mapping[str, str])


class SyncTranslator(BaseTranslator):
    def __init__(
        self,
        proxies: Optional[Dict[str, str]] = None,
        url: str = DEFAULT_TRANSLATION_ENDPOINT,
        tts_url: str = DEFAULT_TTS_ENDPOINT,
        headers: Optional[Union[Dict[str, str], Callable[[], Dict[str, str]]]] = None,
        **options: Any,
    ) -> None:
        self.url = url
        self.tts_url = tts_url
        self.proxies = proxies
        self.options = options
        self.headers = get_base_headers if headers is None else headers

    @overload
    def translate(
        self,
        text: str,
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> TranslatedObject: ...

    @overload
    def translate(
        self,
        text: List[str],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> List[TranslatedObject]: ...

    @overload
    def translate(
        self,
        text: Dict[Any, str],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> Dict[str, TranslatedObject]: ...

    def translate(
        self,
        text: Union[str, List[str], Dict[Any, str], Mapping[str, str]],
        sourcelang: str = "auto",
        targetlang: str = "en",
        client: str = "gtx",
        dt: str = "t",
        dj: int = 1,
        **extra: Any,
    ) -> Union[TranslatedObject, List[TranslatedObject], Dict[str, TranslatedObject]]:
        """
        A function that translates text.

        Parameters:
            text: str = Text to translate.


            sourcelang: str = "text" original language, pass "auto" for auto-detection (default: "auto").


            targetlang: str = target language for translating "text" (default: "en").


            client: str = Google Translate client platform I guess, it's not well documented. (default: "gtx")
                (Will return the raw API json if "client" is not "gtx")


            dt: str = Specifies what to return to the API reply (default: "t").
                (Will return the raw API json if "dt" is not "t".
                t - Translation of source text;
                at - Alternate translations;
                rm - Transcription / transliteration of source and translated texts;
                bd - Dictionary, in case source text is one word
                     (you get translations with articles, reverse translations, etc.);
                md - Definitions of source text, if it's one word;
                ss - Synonyms of source text, if it's one word;
                ex - Examples;
                rw - See also list;
                dj - Json response with names. (dj=1).)

            dj: int = Probably specifies dictionary output if true, else list (default: 1).

            extra: dict = Extra params to be sent to api endpoint.

        Returns:
            TranslatedObject or JSON serialized object.
        """

        params = {
            k: v
            for k, v in {
                "client": client,
                "sl": sourcelang,
                "tl": targetlang,
                "dt": dt,
                "ie": "utf-8",
                "oe": "utf-8",
                "dj": dj,
                **extra,
            }.items()
            if v is not None
        }

        try:
            proxies: Dict[str, httpx.HTTPTransport] = {}
            if self.proxies:
                if proxies.get("https"):
                    proxies["https"] = httpx.HTTPTransport(proxy=self.proxies["https"])
                if proxies.get("http"):
                    proxies["http"] = httpx.HTTPTransport(proxy=self.proxies["http"])
                if proxies.get("socks5"):
                    proxies["socks5"] = httpx.HTTPTransport(proxy=self.proxies["socks5"])
                if proxies.get("socks5h"):
                    proxies["socks5h"] = httpx.HTTPTransport(proxy=self.proxies["socks5h"])

            with httpx.Client(mounts=proxies, **self.options) as c:
                raw: Union[Mapping[str, Any], List[Any]] = (
                    c.post(
                        self.url,
                        params={**params, "q": text},
                        headers=self.get_headers(),
                    ).json()
                    if isinstance(text, str)
                    else (
                        {
                            k: c.post(
                                self.url,
                                params={**params, "q": v},
                                headers=self.get_headers(),
                            ).json()
                            for k, v in text.items()
                        }
                        if isinstance(text, Mapping)
                        else [
                            c.post(
                                self.url,
                                params={**params, "q": elem},
                                headers=self.get_headers(),
                            ).json()
                            for elem in text
                        ]
                    )
                )
                c.close()

            return self.check(raw=raw, client=client, dt=dt, text=text)
        except Exception as e:
            raise TranslationError(e) from None

    def detect(
        self, text: Union[str, List[Any], Dict[Any, Any]], **kwargs: Any
    ) -> Union[str, List[str], Dict[Any, str]]:
        if isinstance(text, str):
            return self(text).lang
        elif isinstance(text, list):
            return [self(elem).lang for elem in text]
        elif isinstance(text, dict):
            return {k: self(v).lang for k, v in text.items()}
        else:
            raise ValueError("Language detection works only with str, list and dict")

    def tts(
        self,
        text: Union[str, List[str], Dict[Any, str], Mapping[str, str]],
        file: BinaryIO,
        targetlang: str = "en",
        client: str = "at",
        idx: int = 0,
        prev: str = "input",
        chunk_size: int = 1024,
        textlen: Optional[int] = None,
        **extra: Any,
    ) -> BinaryIO:
        params = self.parse_tts(
            client=client,
            targetlang=targetlang,
            idx=idx,
            prev=prev,
            text=text,
            textlen=textlen,
            extra=extra,
        )
        try:
            proxies: Dict[str, httpx.HTTPTransport] = {}
            if self.proxies:
                if proxies.get("https"):
                    proxies["https"] = httpx.HTTPTransport(proxy=self.proxies["https"])
                if proxies.get("http"):
                    proxies["http"] = httpx.HTTPTransport(proxy=self.proxies["http"])
                if proxies.get("socks5"):
                    proxies["socks5"] = httpx.HTTPTransport(proxy=self.proxies["socks5"])
                if proxies.get("socks5h"):
                    proxies["socks5h"] = httpx.HTTPTransport(proxy=self.proxies["socks5h"])

            with httpx.stream(
                "GET",
                url=self.tts_url,
                params=params,
                headers=self.get_headers(),
            ) as response:
                for chunk in response.iter_bytes(chunk_size=chunk_size):
                    file.write(chunk)
            return file
        except Exception as e:
            raise TranslationError(e) from None

    __call__ = translate



"""Custom exceptions for gpytranslate."""

from typing import Tuple


class GpytranslateException(Exception):
    """Base exception for all gpytranslate errors."""


class TranslationError(GpytranslateException):
    """Raised when a translation or TTS operation fails."""

    def __init__(self, exception: Exception) -> None:
        """Initialize the error with the underlying exception.

        Args:
            exception: The original exception that caused this error
        """
        self.args: Tuple[Exception, ...] = (exception,)

    def __str__(self) -> str:
        """Get string representation of the error.

        Returns:
            str: Formatted error message with the underlying exception
        """
        return f"Translation/TTS operation failed: {self.args[0]!r}"
